package java.util.prefs;

import java.util.LinkedList;
import java.io.IOException;
import java.io.OutputStream;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.TreeSet;
import java.util.Objects;
import java.util.HashMap;
import java.util.EventObject;
import java.util.List;
import java.util.Map;

public abstract class AbstractPreferences extends Preferences
{
    private final String name;
    private final String absolutePath;
    final AbstractPreferences parent;
    private final AbstractPreferences root;
    protected boolean newNode;
    private Map<String, AbstractPreferences> kidCache;
    private boolean removed;
    private PreferenceChangeListener[] prefListeners;
    private NodeChangeListener[] nodeListeners;
    protected final Object lock;
    private static final String[] EMPTY_STRING_ARRAY;
    private static final AbstractPreferences[] EMPTY_ABSTRACT_PREFS_ARRAY;
    private static final List<EventObject> eventQueue;
    private static Thread eventDispatchThread;
    
    protected AbstractPreferences(AbstractPreferences parent, String name) {
        this.newNode = false;
        this.kidCache = new HashMap();
        this.removed = false;
        this.prefListeners = new PreferenceChangeListener[0];
        this.nodeListeners = new NodeChangeListener[0];
        this.lock = new Object();
        if (parent == null) {
            if (!name.equals("")) {
                throw new IllegalArgumentException("Root name '" + name + "' must be \"\"");
            }
            this.absolutePath = "/";
            this.root = this;
        }
        else {
            if (name.indexOf(47) != -1) {
                throw new IllegalArgumentException("Name '" + name + "' contains '/'");
            }
            if (name.equals("")) {
                throw new IllegalArgumentException("Illegal name: empty string");
            }
            this.root = parent.root;
            this.absolutePath = ((parent == this.root) ? ("/" + name) : (parent.absolutePath() + "/" + name));
        }
        this.name = name;
        this.parent = parent;
    }
    
    @Override
    public void put(String str, String str2) {
        if (str == null || str2 == null) {
            throw new NullPointerException();
        }
        if (str.length() > 80) {
            throw new IllegalArgumentException("Key too long: " + str);
        }
        if (str2.length() > 8192) {
            throw new IllegalArgumentException("Value too long: " + str2);
        }
        synchronized (this.lock) {
            if (this.removed) {
                throw new IllegalStateException("Node has been removed.");
            }
            this.putSpi(str, str2);
            this.enqueuePreferenceChangeEvent(str, str2);
        }
        return;
    }
    
    @Override
    public String get(String s, String s2) {
        if (s == null) {
            throw new NullPointerException("Null key");
        }
        synchronized (this.lock) {
            if (this.removed) {
                throw new IllegalStateException("Node has been removed.");
            }
            String spi = null;
            try {
                spi = this.getSpi(s);
            }
            catch (Exception ex) {}
            return (spi == null) ? s2 : spi;
        }
    }
    
    @Override
    public void remove(String obj) {
        Objects.requireNonNull(obj, "Specified key cannot be null");
        synchronized (this.lock) {
            if (this.removed) {
                throw new IllegalStateException("Node has been removed.");
            }
            this.removeSpi(obj);
            this.enqueuePreferenceChangeEvent(obj, null);
        }
    }
    
    @Override
    public void clear() throws BackingStoreException {
        synchronized (this.lock) {
            String[] keys = this.keys();
            for (int i = 0; i < keys.length; ++i) {
                this.remove(keys[i]);
            }
        }
    }
    
    @Override
    public void putInt(String s, int i) {
        this.put(s, Integer.toString(i));
    }
    
    @Override
    public int getInt(String s, int int1) {
        try {
            String value = this.get(s, null);
            if (value != null) {
                int1 = Integer.parseInt(value);
            }
        }
        catch (NumberFormatException ex) {}
        return int1;
    }
    
    @Override
    public void putLong(String s, long i) {
        this.put(s, Long.toString(i));
    }
    
    @Override
    public long getLong(String s, long long1) {
        try {
            String value = this.get(s, null);
            if (value != null) {
                long1 = Long.parseLong(value);
            }
        }
        catch (NumberFormatException ex) {}
        return long1;
    }
    
    @Override
    public void putBoolean(String s, boolean b) {
        this.put(s, String.valueOf(b));
    }
    
    @Override
    public boolean getBoolean(String s, boolean n) {
        String value = this.get(s, null);
        if (value != null) {
            if (value.equalsIgnoreCase("true")) {
                n = 1;
            }
            else if (value.equalsIgnoreCase("false")) {
                n = 0;
            }
        }
        return n;
    }
    
    @Override
    public void putFloat(String s, float f) {
        this.put(s, Float.toString(f));
    }
    
    @Override
    public float getFloat(String s, float float1) {
        try {
            String value = this.get(s, null);
            if (value != null) {
                float1 = Float.parseFloat(value);
            }
        }
        catch (NumberFormatException ex) {}
        return float1;
    }
    
    @Override
    public void putDouble(String s, double d) {
        this.put(s, Double.toString(d));
    }
    
    @Override
    public double getDouble(String s, double double1) {
        try {
            String value = this.get(s, null);
            if (value != null) {
                double1 = Double.parseDouble(value);
            }
        }
        catch (NumberFormatException ex) {}
        return double1;
    }
    
    @Override
    public void putByteArray(String s, byte[] array) {
        this.put(s, Base64.byteArrayToBase64(array));
    }
    
    @Override
    public byte[] getByteArray(String s, byte[] base64ToByteArray) {
        String value = this.get(s, null);
        try {
            if (value != null) {
                base64ToByteArray = Base64.base64ToByteArray(value);
            }
        }
        catch (RuntimeException ex) {}
        return base64ToByteArray;
    }
    
    @Override
    public String[] keys() throws BackingStoreException {
        synchronized (this.lock) {
            if (this.removed) {
                throw new IllegalStateException("Node has been removed.");
            }
            return this.keysSpi();
        }
    }
    
    @Override
    public String[] childrenNames() throws BackingStoreException {
        synchronized (this.lock) {
            if (this.removed) {
                throw new IllegalStateException("Node has been removed.");
            }
            TreeSet set = new TreeSet(this.kidCache.keySet());
            String[] childrenNamesSpi = this.childrenNamesSpi();
            for (int length = childrenNamesSpi.length, i = 0; i < length; ++i) {
                set.add(childrenNamesSpi[i]);
            }
            return (String[])set.toArray(AbstractPreferences.EMPTY_STRING_ARRAY);
        }
    }
    
    protected final AbstractPreferences[] cachedChildren() {
        return (AbstractPreferences[])this.kidCache.values().toArray(AbstractPreferences.EMPTY_ABSTRACT_PREFS_ARRAY);
    }
    
    @Override
    public Preferences parent() {
        synchronized (this.lock) {
            if (this.removed) {
                throw new IllegalStateException("Node has been removed.");
            }
            return this.parent;
        }
    }
    
    @Override
    public Preferences node(String str) {
        synchronized (this.lock) {
            if (this.removed) {
                throw new IllegalStateException("Node has been removed.");
            }
            if (str.equals("")) {
                return this;
            }
            if (str.equals("/")) {
                return this.root;
            }
            if (str.charAt(0) != '/') {
                return this.node(new StringTokenizer(str, "/", true));
            }
        }
        return this.root.node(new StringTokenizer(str.substring(1), "/", true));
    }
    
    private Preferences node(StringTokenizer stringTokenizer) {
        String nextToken = stringTokenizer.nextToken();
        if (nextToken.equals("/")) {
            throw new IllegalArgumentException("Consecutive slashes in path");
        }
        synchronized (this.lock) {
            AbstractPreferences childSpi = (AbstractPreferences)this.kidCache.get(nextToken);
            if (childSpi == null) {
                if (nextToken.length() > 80) {
                    throw new IllegalArgumentException("Node name " + nextToken + " too long");
                }
                childSpi = this.childSpi(nextToken);
                if (childSpi.newNode) {
                    this.enqueueNodeAddedEvent(childSpi);
                }
                this.kidCache.put(nextToken, childSpi);
            }
            if (!stringTokenizer.hasMoreTokens()) {
                return childSpi;
            }
            stringTokenizer.nextToken();
            if (!stringTokenizer.hasMoreTokens()) {
                throw new IllegalArgumentException("Path ends with slash");
            }
            return childSpi.node(stringTokenizer);
        }
    }
    
    @Override
    public boolean nodeExists(String str) throws BackingStoreException {
        synchronized (this.lock) {
            if (str.equals("")) {
                return !this.removed;
            }
            if (this.removed) {
                throw new IllegalStateException("Node has been removed.");
            }
            if (str.equals("/")) {
                return true;
            }
            if (str.charAt(0) != '/') {
                return this.nodeExists(new StringTokenizer(str, "/", true));
            }
        }
        return this.root.nodeExists(new StringTokenizer(str.substring(1), "/", true));
    }
    
    private boolean nodeExists(StringTokenizer stringTokenizer) throws BackingStoreException {
        String nextToken = stringTokenizer.nextToken();
        if (nextToken.equals("/")) {
            throw new IllegalArgumentException("Consecutive slashes in path");
        }
        synchronized (this.lock) {
            AbstractPreferences child = (AbstractPreferences)this.kidCache.get(nextToken);
            if (child == null) {
                child = this.getChild(nextToken);
            }
            if (child == null) {
                return false;
            }
            if (!stringTokenizer.hasMoreTokens()) {
                return true;
            }
            stringTokenizer.nextToken();
            if (!stringTokenizer.hasMoreTokens()) {
                throw new IllegalArgumentException("Path ends with slash");
            }
            return child.nodeExists(stringTokenizer);
        }
    }
    
    @Override
    public void removeNode() throws BackingStoreException {
        if (this == this.root) {
            throw new UnsupportedOperationException("Can't remove the root!");
        }
        synchronized (this.parent.lock) {
            this.removeNode2();
            this.parent.kidCache.remove(this.name);
        }
        return;
    }
    
    private void removeNode2() throws BackingStoreException {
        synchronized (this.lock) {
            if (this.removed) {
                throw new IllegalStateException("Node already removed.");
            }
            String[] childrenNamesSpi = this.childrenNamesSpi();
            for (int i = 0; i < childrenNamesSpi.length; ++i) {
                if (!this.kidCache.containsKey(childrenNamesSpi[i])) {
                    this.kidCache.put(childrenNamesSpi[i], this.childSpi(childrenNamesSpi[i]));
                }
            }
            Iterator<AbstractPreferences> iterator = this.kidCache.values().iterator();
            while (iterator.hasNext()) {
                try {
                    ((AbstractPreferences)iterator.next()).removeNode2();
                    iterator.remove();
                }
                catch (BackingStoreException ex) {}
            }
            this.removeNodeSpi();
            this.removed = true;
            this.parent.enqueueNodeRemovedEvent(this);
        }
    }
    
    @Override
    public String name() {
        return this.name;
    }
    
    @Override
    public String absolutePath() {
        return this.absolutePath;
    }
    
    @Override
    public boolean isUserNode() {
        return (boolean)AccessController.doPrivileged(new PrivilegedAction<Boolean>() {
            @Override
            public Boolean run() {
                return Boolean.valueOf(AbstractPreferences.this.root == Preferences.userRoot());
            }
        });
    }
    
    @Override
    public void addPreferenceChangeListener(PreferenceChangeListener preferenceChangeListener) {
        if (preferenceChangeListener == null) {
            throw new NullPointerException("Change listener is null.");
        }
        synchronized (this.lock) {
            if (this.removed) {
                throw new IllegalStateException("Node has been removed.");
            }
            PreferenceChangeListener[] prefListeners = this.prefListeners;
            System.arraycopy(prefListeners, 0, this.prefListeners = new PreferenceChangeListener[prefListeners.length + 1], 0, prefListeners.length);
            this.prefListeners[prefListeners.length] = preferenceChangeListener;
        }
        startEventDispatchThreadIfNecessary();
        return;
    }
    
    @Override
    public void removePreferenceChangeListener(PreferenceChangeListener preferenceChangeListener) {
        synchronized (this.lock) {
            if (this.removed) {
                throw new IllegalStateException("Node has been removed.");
            }
            if (this.prefListeners == null || this.prefListeners.length == 0) {
                throw new IllegalArgumentException("Listener not registered.");
            }
            PreferenceChangeListener[] prefListeners;
            int i;
            for (prefListeners = new PreferenceChangeListener[this.prefListeners.length - 1], i = 0; i < prefListeners.length && this.prefListeners[i] != preferenceChangeListener; prefListeners[i] = this.prefListeners[i++]) {}
            if (i == prefListeners.length && this.prefListeners[i] != preferenceChangeListener) {
                throw new IllegalArgumentException("Listener not registered.");
            }
            while (i < prefListeners.length) {
                prefListeners[i] = this.prefListeners[++i];
            }
            this.prefListeners = prefListeners;
        }
    }
    
    @Override
    public void addNodeChangeListener(NodeChangeListener nodeChangeListener) {
        if (nodeChangeListener == null) {
            throw new NullPointerException("Change listener is null.");
        }
        synchronized (this.lock) {
            if (this.removed) {
                throw new IllegalStateException("Node has been removed.");
            }
            if (this.nodeListeners == null) {
                (this.nodeListeners = new NodeChangeListener[1])[0] = nodeChangeListener;
            }
            else {
                NodeChangeListener[] nodeListeners = this.nodeListeners;
                System.arraycopy(nodeListeners, 0, this.nodeListeners = new NodeChangeListener[nodeListeners.length + 1], 0, nodeListeners.length);
                this.nodeListeners[nodeListeners.length] = nodeChangeListener;
            }
        }
        startEventDispatchThreadIfNecessary();
        return;
    }
    
    @Override
    public void removeNodeChangeListener(NodeChangeListener nodeChangeListener) {
        synchronized (this.lock) {
            if (this.removed) {
                throw new IllegalStateException("Node has been removed.");
            }
            if (this.nodeListeners == null || this.nodeListeners.length == 0) {
                throw new IllegalArgumentException("Listener not registered.");
            }
            int n;
            for (n = 0; n < this.nodeListeners.length && this.nodeListeners[n] != nodeChangeListener; ++n) {}
            if (n == this.nodeListeners.length) {
                throw new IllegalArgumentException("Listener not registered.");
            }
            NodeChangeListener[] nodeListeners = new NodeChangeListener[this.nodeListeners.length - 1];
            if (n != 0) {
                System.arraycopy(this.nodeListeners, 0, nodeListeners, 0, n);
            }
            if (n != nodeListeners.length) {
                System.arraycopy(this.nodeListeners, n + 1, nodeListeners, n, nodeListeners.length - n);
            }
            this.nodeListeners = nodeListeners;
        }
    }
    
    protected abstract void putSpi(String p0, String p1);
    
    protected abstract String getSpi(String p0);
    
    protected abstract void removeSpi(String p0);
    
    protected abstract void removeNodeSpi() throws BackingStoreException;
    
    protected abstract String[] keysSpi() throws BackingStoreException;
    
    protected abstract String[] childrenNamesSpi() throws BackingStoreException;
    
    protected AbstractPreferences getChild(String anObject) throws BackingStoreException {
        synchronized (this.lock) {
            String[] childrenNames = this.childrenNames();
            int i = 0;
            while (i < childrenNames.length) {
                if (childrenNames[i].equals(anObject)) {
                    return this.childSpi(childrenNames[i]);
                }
                ++i;
            }
        }
        return null;
    }
    
    protected abstract AbstractPreferences childSpi(String p0);
    
    @Override
    public String toString() {
        return (this.isUserNode() ? "User" : "System") + " Preference Node: " + this.absolutePath();
    }
    
    @Override
    public void sync() throws BackingStoreException {
        this.sync2();
    }
    
    private void sync2() throws BackingStoreException {
        AbstractPreferences[] cachedChildren;
        synchronized (this.lock) {
            if (this.removed) {
                throw new IllegalStateException("Node has been removed");
            }
            this.syncSpi();
            cachedChildren = this.cachedChildren();
        }
        for (int i = 0; i < cachedChildren.length; ++i) {
            cachedChildren[i].sync2();
        }
    }
    
    protected abstract void syncSpi() throws BackingStoreException;
    
    @Override
    public void flush() throws BackingStoreException {
        this.flush2();
    }
    
    private void flush2() throws BackingStoreException {
        AbstractPreferences[] cachedChildren;
        synchronized (this.lock) {
            this.flushSpi();
            if (this.removed) {
                return;
            }
            cachedChildren = this.cachedChildren();
        }
        for (int i = 0; i < cachedChildren.length; ++i) {
            cachedChildren[i].flush2();
        }
    }
    
    protected abstract void flushSpi() throws BackingStoreException;
    
    protected boolean isRemoved() {
        synchronized (this.lock) {
            return this.removed;
        }
    }
    
    private static synchronized void startEventDispatchThreadIfNecessary() {
        if (AbstractPreferences.eventDispatchThread == null) {
            (AbstractPreferences.eventDispatchThread = new AbstractPreferences.EventDispatchThread()).setDaemon(true);
            AbstractPreferences.eventDispatchThread.start();
        }
    }
    
    PreferenceChangeListener[] prefListeners() {
        synchronized (this.lock) {
            return this.prefListeners;
        }
    }
    
    NodeChangeListener[] nodeListeners() {
        synchronized (this.lock) {
            return this.nodeListeners;
        }
    }
    
    private void enqueuePreferenceChangeEvent(String s, String s2) {
        if (this.prefListeners.length != 0) {
            synchronized (AbstractPreferences.eventQueue) {
                AbstractPreferences.eventQueue.add(new PreferenceChangeEvent(this, s, s2));
                AbstractPreferences.eventQueue.notify();
            }
        }
    }
    
    private void enqueueNodeAddedEvent(Preferences preferences) {
        if (this.nodeListeners.length != 0) {
            synchronized (AbstractPreferences.eventQueue) {
                AbstractPreferences.eventQueue.add(new AbstractPreferences.NodeAddedEvent(this, preferences));
                AbstractPreferences.eventQueue.notify();
            }
        }
    }
    
    private void enqueueNodeRemovedEvent(Preferences preferences) {
        if (this.nodeListeners.length != 0) {
            synchronized (AbstractPreferences.eventQueue) {
                AbstractPreferences.eventQueue.add(new AbstractPreferences.NodeRemovedEvent(this, preferences));
                AbstractPreferences.eventQueue.notify();
            }
        }
    }
    
    @Override
    public void exportNode(OutputStream outputStream) throws IOException, BackingStoreException {
        XmlSupport.export(outputStream, this, false);
    }
    
    @Override
    public void exportSubtree(OutputStream outputStream) throws IOException, BackingStoreException {
        XmlSupport.export(outputStream, this, true);
    }
    
    static {
        EMPTY_STRING_ARRAY = new String[0];
        EMPTY_ABSTRACT_PREFS_ARRAY = new AbstractPreferences[0];
        eventQueue = new LinkedList();
        AbstractPreferences.eventDispatchThread = null;
    }
    
    private class NodeAddedEvent extends NodeChangeEvent
    {
        private static final long serialVersionUID = -6743557530157328528L;
        
        NodeAddedEvent(Preferences preferences, Preferences preferences2) {
            super(preferences, preferences2);
        }
    }
    
    private class NodeRemovedEvent extends NodeChangeEvent
    {
        private static final long serialVersionUID = 8735497392918824837L;
        
        NodeRemovedEvent(Preferences preferences, Preferences preferences2) {
            super(preferences, preferences2);
        }
    }
    
    private static class EventDispatchThread extends Thread
    {
        @Override
        public void run() {
            while (true) {
                EventObject eventObject = null;
                synchronized (AbstractPreferences.eventQueue) {
                    try {
                        while (AbstractPreferences.eventQueue.isEmpty()) {
                            AbstractPreferences.eventQueue.wait();
                        }
                        eventObject = (EventObject)AbstractPreferences.eventQueue.remove(0);
                    }
                    catch (InterruptedException ex) {
                        return;
                    }
                }
                AbstractPreferences abstractPreferences = (AbstractPreferences)eventObject.getSource();
                if ((eventObject instanceof PreferenceChangeEvent)) {
                    PreferenceChangeEvent preferenceChangeEvent = (PreferenceChangeEvent)eventObject;
                    PreferenceChangeListener[] prefListeners = abstractPreferences.prefListeners();
                    for (int k = 0; k < prefListeners.length; ++k) {
                        prefListeners[k].preferenceChange(preferenceChangeEvent);
                    }
                }
                else {
                    NodeChangeEvent nodeChangeEvent = (NodeChangeEvent)eventObject;
                    NodeChangeListener[] nodeListeners = abstractPreferences.nodeListeners();
                    if ((nodeChangeEvent instanceof AbstractPreferences.NodeAddedEvent)) {
                        for (int j = 0; j < nodeListeners.length; ++j) {
                            nodeListeners[j].childAdded(nodeChangeEvent);
                        }
                    }
                    else {
                        for (int i = 0; i < nodeListeners.length; ++i) {
                            nodeListeners[i].childRemoved(nodeChangeEvent);
                        }
                    }
                }
            }
        }
    }
}
